/*
 * iSCSI connection daemon
 * Copyright (C) 2002 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-discovery.c,v 1.5.2.5 2004/09/16 09:08:37 surekhap Exp $
 *
 */

#include <stdlib.h>
#include <stdarg.h>

#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <sys/time.h>

#include "iscsi-protocol.h"
#include "iscsi-platform.h"
#include "iscsi-session.h"
#include "iscsi-io.h"
#include "iscsi-login.h"
#include "iscsi-hooks.h"
#include "iscsi-bindings.h"
#include "iscsi-discovery.h"
#include "string-buffer.h"

#ifdef SLP_ENABLE
#include "iscsi-slp-discovery.h"
#endif

static volatile int rediscover = 0;
static int record_begin;

static void
sighup_handler(int unused)
{
    rediscover = 1;
}

int
send_nop_reply(iscsi_session_t * session, struct IscsiNopInHdr *nop, char *data,
	       int timeout)
{
    struct IscsiNopOutHdr out;

    memset(&out, 0, sizeof (out));
    out.opcode = ISCSI_OP_NOOP_OUT | ISCSI_OP_IMMEDIATE;
    out.flags = ISCSI_FLAG_FINAL;
    memcpy(out.lun, nop->lun, sizeof (out.lun));
    out.itt = nop->itt;
    out.ttt = nop->ttt;
    memcpy(out.dlength, nop->dlength, sizeof (out.dlength));
    out.cmdsn = htonl(session->CmdSn);	/* don't increment after immediate cmds */
    out.expstatsn = htonl(session->ExpStatSn);

    debugmsg(4, "sending nop reply for ttt %u, cmdsn %u, dlength %d",
	     ntohl(out.ttt), ntohl(out.cmdsn), ntoh24(out.dlength));

    return iscsi_send_pdu(session, (struct IscsiHdr *) &out, data, timeout);
}

int
request_targets(iscsi_session_t * session)
{
    char data[64];
    struct IscsiTextHdr text;
    struct IscsiHdr *pdu = (struct IscsiHdr *) &text;

    memset(&text, 0, sizeof (text));
    memset(data, 0, sizeof (data));

    /* make a text PDU with SendTargets=All */
    if (!iscsi_make_text_pdu(session, pdu, data, sizeof (data))) {
	logmsg(AS_ERROR, "failed to make a SendTargets PDU");
	return 0;
    }

    if (!iscsi_add_text
	(session, pdu, data, sizeof (data), "SendTargets", "All")) {
	logmsg(AS_ERROR, "failed to add SendTargets text key");
	exit(1);
    }

    text.ttt = RSVD_TASK_TAG;
    text.flags = ISCSI_FLAG_FINAL;

    if (++session->itt == RSVD_TASK_TAG)
	session->itt = 1;

    if (!iscsi_send_pdu(session, pdu, data, session->active_timeout)) {
	logmsg(AS_ERROR, "failed to send SendTargets PDU");
	return 0;
    }

    return 1;
}

int
iterate_targets(iscsi_session_t * session, uint32_t ttt)
{
    char data[64];
    struct IscsiTextHdr text;
    struct IscsiHdr *pdu = (struct IscsiHdr *) &text;

    memset(&text, 0, sizeof (text));
    memset(data, 0, sizeof (data));

    /* make an empty text PDU */
    if (!iscsi_make_text_pdu(session, pdu, data, sizeof (data))) {
	logmsg(AS_ERROR, "failed to make an empty text PDU");
	return 0;
    }

    text.ttt = ttt;
    text.flags = ISCSI_FLAG_FINAL;

    if (++session->itt == RSVD_TASK_TAG)
	session->itt = 1;

    if (!iscsi_send_pdu(session, pdu, data, session->active_timeout)) {
	logmsg(AS_ERROR, "failed to send empty text PDU");
	return 0;
    }

    return 1;
}

/* try to reset the session's IP address and port, based on
 * the TargetAddress provided 
 */
static int
update_address(iscsi_session_t * session, char *address)
{
    char *port;
    char *tag;
    struct hostent *hostn = NULL;

    if ((tag = iscsi_strrchr(address, ','))) {
	*tag = '\0';
	tag++;
    }
    if ((port = iscsi_strrchr(address, ':'))) {
	*port = '\0';
	port++;
    }

    do {
	hostn = gethostbyname(address);
	if (hostn == NULL) {
	    errormsg("cannot resolve host name %s", address);
	    sleep(1);
	}
	if (iscsi_process_should_exit())
	    return 0;
    } while (!hostn);

    memcpy(session->ip_address, hostn->h_addr,
	   MIN(sizeof (session->ip_address), hostn->h_length));
    session->ip_length = hostn->h_length;

    if (port) {
	session->port = iscsi_atoi(port);
    } else {
	session->port = ISCSI_LISTEN_PORT;
    }

    return 1;
}

int
add_portal(string_buffer_t * info, char *address, char *port, char *tag)
{
    struct hostent *hostn = NULL;

    /* resolve the address, in case it was a DNS name */
    do {
	hostn = gethostbyname(address);
	if (!hostn) {
	    errormsg("cannot resolve %s", address);
	    sleep(1);
	}
	if (iscsi_process_should_exit())
	    return 0;
    } while (!hostn);

    /* convert the resolved name to text */
    if (hostn->h_length == 4) {
	struct in_addr addr;

	memcpy(&addr, hostn->h_addr, sizeof (addr));

	if (tag && *tag) {
	    if (!append_sprintf(info, "TT=%s\n", tag)) {
		logmsg(AS_ERROR, "couldn't add portal tag %s", tag);
		return 0;
	    }
	}

	if (port && *port) {
	    if (!append_sprintf(info, "TP=%s\n", port)) {
		logmsg(AS_ERROR, "couldn't add port %s", port);
		return 0;
	    }
	}

	if (strcmp(inet_ntoa(addr), address)) {
	    /* if the resolved name doesn't match the
	     * original, send an RA line as well as a TA
	     * line 
	     */
	    return append_sprintf(info, "RA=%s\nTA=%s\n", inet_ntoa(addr),
				  address);
	} else {
	    /* don't need the RA line */
	    return append_sprintf(info, "TA=%s\n", address);
	}
    } else {
	/* FIXME: IPv6 */
	logmsg(AS_ERROR, "can't handle network address %s", address);
	return 0;
    }
}

int
add_target_record(string_buffer_t * info, char *name, char *end,
		  int lun_inventory_changed, char *default_address,
		  char *default_port, int fd)
{
    char *text = NULL;
    char *nul = name;
    size_t length;
    size_t original = data_length(info);

    /* address = IPv4
     * address = [IPv6]
     * address = DNSname
     * address = IPv4:port
     * address = [IPv6]:port
     * address = DNSname:port
     * address = IPv4,tag
     * address = [IPv6],tag
     * address = DNSname,tag
     * address = IPv4:port,tag
     * address = [IPv6]:port,tag
     * address = DNSname:port,tag
     */

    debugmsg(7, "adding target record %p, end %p", name, end);

    /* find the end of the name */
    while ((nul < end) && (*nul != '\0'))
	nul++;

    length = nul - name;
    if (length > TARGET_NAME_MAXLEN) {
	logmsg(AS_ERROR, "TargetName %s too long, ignoring", name);
	return 0;
    }

    if (!record_begin) {
	if (!append_sprintf
	    (info, lun_inventory_changed ? "DLC=%s\n" : "DTN=%s\n", name)) {
	    logmsg(AS_ERROR, "couldn't report target %s", name);
	    truncate_buffer(info, original);
	    return 0;
	}
	record_begin = 1;
    } else {
	if (!append_sprintf
	    (info, lun_inventory_changed ? "LC=%s\n" : "TN=%s\n", name)) {
	    logmsg(AS_ERROR, "couldn't report target %s", name);
	    truncate_buffer(info, original);
	    return 0;
	}
    }

    text = name + length;

    /* skip NULs after the name */
    while ((text < end) && (*text == '\0'))
	text++;

    /* if no address is provided, use the default */
    if (text >= end) {
	if (default_address == NULL) {
	    logmsg(AS_ERROR, "no default address known for target %s", name);
	    truncate_buffer(info, original);
	    return 0;
	} else if (!add_portal(info, default_address, default_port, NULL)) {
	    logmsg(AS_ERROR, "failed to add default portal, ignoring target %s",
		   name);
	    truncate_buffer(info, original);
	    return 0;
	} else if (!append_string(info, ";\n")) {
	    logmsg(AS_ERROR,
		   "failed to terminate target record, ignoring target %s",
		   name);
	    truncate_buffer(info, original);
	    return 0;
	}

	/* finished adding the default */
	return 1;
    }

    /* process TargetAddresses */
    while (text < end) {
	char *next = text + strlen(text) + 1;

	debugmsg(7, "text %p, next %p, end %p, %s", text, next, end, text);

	if (iscsi_strncmp(text, "TargetAddress=", 14) == 0) {
	    char *port = NULL;
	    char *tag = NULL;
	    char *address = text + 14;

	    /* FIXME: handle IPv6 */
	    if (address[0] == '[') {
		/* This is an IPv6 numeric address; skip it */
		text = next;
		continue;
	    }
	    if ((tag = strrchr(text, ','))) {
		*tag = '\0';
		tag++;
	    }
	    if ((port = strrchr(text, ':'))) {
		*port = '\0';
		port++;
	    }

	    if (!add_portal(info, address, port, tag)) {
		logmsg(AS_ERROR,
		       "failed to add default portal, ignoring target %s",
		       name);
		truncate_buffer(info, original);
		return 0;
	    }
	} else {
	    logmsg(AS_ERROR, "unexpected SendTargets data: %s", text);
	}

	text = next;
    }

    /* indicate the end of the target record */
    if (!append_string(info, ";\n")) {
	logmsg(AS_ERROR,
	       "failed to terminate target record, ignoring target %s", name);
	truncate_buffer(info, original);
	return 0;
    }

    return 1;
}

static int
process_sendtargets_response(string_buffer_t * sendtargets,
			     string_buffer_t * info, int final,
			     int lun_inventory_changed, char *default_address,
			     char *default_port, int fd)
{
    char *start = buffer_data(sendtargets);
    char *text = start;
    char *end = text + data_length(sendtargets);
    char *nul = end - 1;
    char *record = NULL;
    int num_targets = 0;
    size_t valid_info = 0;

    if (start == end) {
	/* no SendTargets data */
	goto done;
    }

    /* scan backwards to find the last NUL in the data, to ensure we
     * don't walk off the end.  Since key=value pairs can span PDU
     * boundaries, we're not guaranteed that the end of the data has a
     * NUL.
     */
    while ((nul > start) && *nul)
	nul--;

    if (nul == start) {
	/* couldn't find anything we can process now, it's one big partial string */
	goto done;
    }

    /* find the boundaries between target records (TargetName or final PDU) */
    for (;;) {
	/* skip NULs */
	while ((text < nul) && (*text == '\0'))
	    text++;

	if (text == nul)
	    break;

	debugmsg(7, "processing sendtargets record %p, text %p, line %s",
		 record, text, text);

	/* look for the start of a new target record */
	if (iscsi_strncmp(text, "TargetName=", 11) == 0) {
	    if (record) {
		/* send the last record, which we just found the end of.
		 * don't bother passing the "TargetName=" prefix.
		 */
		if (!add_target_record
		    (info, record + 11, text, lun_inventory_changed,
		     default_address, default_port, fd)) {
		    logmsg(AS_ERROR, "failed to add target record");
		    truncate_buffer(sendtargets, 0);
		    truncate_buffer(info, valid_info);
		    goto done;
		}
		num_targets++;
		valid_info = data_length(info);
	    }
	    record = text;
	}

	/* everything up til the next NUL must be part of
	 * the current target record 
	 */
	while ((text < nul) && (*text != '\0'))
	    text++;
    }

    if (record) {
	if (final) {
	    /* if this is the last PDU of the text sequence,
	     * it also ends a target record 
	     */
	    debugmsg(7, "processing final sendtargets record %p, line %s",
		     record, record);
	    if (add_target_record
		(info, record + 11, text, lun_inventory_changed,
		 default_address, default_port, fd)) {
		num_targets++;
		record = NULL;
		truncate_buffer(sendtargets, 0);
	    } else {
		logmsg(AS_ERROR, "failed to add target record");
		truncate_buffer(sendtargets, 0);
		truncate_buffer(info, valid_info);
		goto done;
	    }
	} else {
	    /* remove the parts of the sendtargets buffer we've
	     * processed, and move the parts we haven't to the
	     * beginning of the buffer.
	     */
	    debugmsg(7, "processed %d bytes of sendtargets data, %d remaining",
		     record - buffer_data(sendtargets),
		     buffer_data(sendtargets) + data_length(sendtargets) -
		     record);
	    remove_initial(sendtargets, record - buffer_data(sendtargets));
	}
    }

  done:
    /* send all of the discovered targets to the parent daemon */
    if (append_string(info, final ? "!\n" : ".\n")) {
	write_buffer(info, fd);
	truncate_buffer(info, 0);
	if (final) {
	    record_begin = 0;
	}
	debugmsg(4, "sent %d targets to parent daemon", num_targets);
	return 1;
    } else {
	logmsg(AS_ERROR, "couldn't send %d targets to parent", num_targets);
	truncate_buffer(info, 0);
	return 0;
    }

    return 1;
}

int
add_async_record(string_buffer_t * info, char *record, int targetoffline,
		 int fd)
{
    int length = strlen(record);
    size_t original = data_length(info);

    debugmsg(7, " adding async record for %s", record);

    if (targetoffline) {
	/* We have received targetoffline event */
	if (length > TARGET_NAME_MAXLEN) {
	    logmsg(AS_ERROR, "Targetname %s too long, ignoring", record);
	    return 0;
	}
    }
    if (!append_sprintf(info, targetoffline ? "ATF=%s\n" : "APF=%s\n", record)) {
	logmsg(AS_ERROR, "couldn't report the record\n");
	truncate_buffer(info, original);
	return 0;
    } else if (!append_string(info, ";\n")) {
	logmsg(AS_ERROR,
	       "failed to terminate target record, ignoring target %s", record);
	truncate_buffer(info, original);
    }
    return 1;
}

static void
clear_timer(struct timeval *timer)
{
    memset(timer, 0, sizeof (*timer));
}

static int
process_async_event_text(string_buffer_t * sendtargets, string_buffer_t * info,
			 struct timeval *timer, int fd)
{
    char *text = buffer_data(sendtargets);
    int targetoffline = 0;
    int slen = (ntohs(*(short *) (text)));
    text = text + 2 + slen;

    if (iscsi_strncmp(text, "X-com.cisco.targetOffline=", 26) == 0) {
	targetoffline = 1;
	if (!add_async_record(info, text + 26, targetoffline, fd)) {
	    logmsg(AS_ERROR, "failed to add async record");
	    return 0;
	}
	clear_timer(timer);
    } else if (iscsi_strncmp(text, "X-com.cisco.portalOffline=", 26) == 0) {
	targetoffline = 0;
	if (!add_async_record(info, text + 26, targetoffline, fd)) {
	    logmsg(AS_ERROR, "failed to add async record");
	    return 0;
	}
	clear_timer(timer);
    } else {
	debugmsg(1, "sendtargets for the other events\n");
	return 1;
    }

    if (append_string(info, "!\n")) {
	write_buffer(info, fd);
	truncate_buffer(info, 0);
	debugmsg(4, "sent async event record to parent daemon\n");
	return 1;
    } else {
	logmsg(AS_ERROR, "couldn't send async event record \n");
	return 0;
    }
}

/* set timer to now + seconds */
static void
set_timer(struct timeval *timer, int seconds)
{
    if (timer) {
	memset(timer, 0, sizeof (*timer));
	gettimeofday(timer, NULL);

	timer->tv_sec += seconds;
    }
}

static int
timer_expired(struct timeval *timer)
{
    struct timeval now;

    /* no timer, can't have expired */
    if ((timer == NULL) || ((timer->tv_sec == 0) && (timer->tv_usec == 0)))
	return 0;

    memset(&now, 0, sizeof (now));
    gettimeofday(&now, NULL);

    if (now.tv_sec > timer->tv_sec)
	return 1;
    if ((now.tv_sec == timer->tv_sec) && (now.tv_usec >= timer->tv_usec))
	return 1;
    return 0;
}

static int
msecs_until(struct timeval *timer)
{
    struct timeval now;
    int msecs;
    long partial;

    /* no timer, can't have expired, infinite time til it expires */
    if ((timer == NULL) || ((timer->tv_sec == 0) && (timer->tv_usec == 0)))
	return -1;

    memset(&now, 0, sizeof (now));
    gettimeofday(&now, NULL);

    /* already expired? */
    if (now.tv_sec > timer->tv_sec)
	return 0;
    if ((now.tv_sec == timer->tv_sec) && (now.tv_usec >= timer->tv_usec))
	return 0;

    /* not expired yet, do the math */
    partial = timer->tv_usec - now.tv_usec;
    if (partial < 0) {
	partial += 1000 * 1000;
	msecs = (partial + 500) / 1000;
	msecs += (timer->tv_sec - now.tv_sec - 1) * 1000;
    } else {
	msecs = (partial + 500) / 1000;
	msecs += (timer->tv_sec - now.tv_sec) * 1000;
    }

    return msecs;
}

int
soonest_msecs(struct timeval *t1, struct timeval *t2, struct timeval *t3)
{
    int m1 = msecs_until(t1);
    int m2 = msecs_until(t2);
    int m3 = msecs_until(t3);

    /* infinity is -1, handle it specically */
    if ((m1 == -1) && (m2 == -1))
	return m3;

    if ((m1 == -1) && (m3 == -1))
	return m2;

    if ((m2 == -1) && (m3 == -1))
	return m1;

    if (m1 == -1)
	return (m2 < m3) ? m2 : m3;

    if (m2 == -1)
	return (m1 < m3) ? m1 : m3;

    if (m3 == -1)
	return (m1 < m2) ? m1 : m2;

    if (m1 < m2)
	return (m1 < m3) ? m1 : m3;
    else
	return (m2 < m3) ? m2 : m3;
}

static void
sendtargets_discovery_process(iscsi_discovery_process_t * process)
{
    iscsi_sendtargets_config_t *config = process->entry->config.sendtargets;
    struct sigaction action;
    iscsi_session_t *session;
    struct hostent *hostn = NULL;
    struct pollfd pfd;
    struct IscsiHdr pdu_buffer;
    struct IscsiHdr *pdu = &pdu_buffer;
    char *data = NULL;
    char *end_of_data;
    int long_lived = (config->continuous > 0) ? 1 : 0;
    int lun_inventory_changed = 0;
    int active = 0;
    struct timeval connection_timer, async_timer;
    int timeout;
    int rc;
    string_buffer_t sendtargets;
    string_buffer_t info;
    uint8_t status_class = 0, status_detail = 0;
    unsigned int login_failures = 0;
    int login_delay = 0;
    char ip_address[16];
    char default_port[12];
    int ip_length = 0;
    int port = config->port;

    /* initial setup */
    process->pid = getpid();
    debugmsg(1, "sendtargets discovery process %p pid %d starting, "
	     "address %s:%d, continuous %d",
	     process, process->pid, config->address, config->port,
	     config->continuous);

    memset(&pdu_buffer, 0, sizeof (pdu_buffer));
    clear_timer(&connection_timer);
    clear_timer(&async_timer);

    /* set SIGTERM, SIGINT to kill the process */
    memset(&action, 0, sizeof (struct sigaction));
    action.sa_sigaction = NULL;
    action.sa_flags = 0;
    action.sa_handler = SIG_DFL;
    sigaction(SIGTERM, &action, NULL);
    sigaction(SIGINT, &action, NULL);

    /* set SIGPIPE to be ignored, so that we get EPIPE */
    action.sa_handler = SIG_IGN;
    sigaction(SIGPIPE, &action, NULL);

    /* set SIGHUP to request a rediscovery */
    action.sa_handler = sighup_handler;
    sigaction(SIGHUP, &action, NULL);

    /* check if a signal arrived before we reset the handler */
    if (iscsi_process_should_exit()) {
	debugmsg(1, "sendtargets discovery process %p exiting\n", process);
	exit(0);
    }

    /* allocate data buffers for SendTargets data and discovery pipe info */
    init_string_buffer(&sendtargets, 32 * 1024);
    init_string_buffer(&info, 8 * 1024);

    session = calloc(1, sizeof (*session));
    if (session == NULL) {
	logmsg(AS_ERROR,
	       "discovery process to %s:%d failed to allocate a session\n",
	       config->address, config->port);
	exit(1);
    }

    /* initialize the session */
    session->socket_fd = -1;
    session->type = ISCSI_SESSION_TYPE_DISCOVERY;
    session->login_timeout = config->connection_timeout_options.login_timeout;
    session->auth_timeout = config->connection_timeout_options.auth_timeout;
    session->active_timeout = config->connection_timeout_options.active_timeout;
    session->idle_timeout = config->connection_timeout_options.idle_timeout;
    session->ping_timeout = config->connection_timeout_options.ping_timeout;
    session->send_async_text =
	config->continuous ? config->send_async_text : -1;
    /* OUI and uniqifying number */
    session->isid[0] = DRIVER_ISID_0;
    session->isid[1] = DRIVER_ISID_1;
    session->isid[2] = DRIVER_ISID_2;
    session->isid[3] = (process->order >> 16) & 0xFF;
    session->isid[4] = (process->order >> 8) & 0xFF;
    session->isid[5] = (process->order >> 0) & 0xFF;
    session->InitiatorName = daemon_config.initiator_name;
    session->InitiatorAlias = daemon_config.initiator_alias;
    session->HeaderDigest = ISCSI_DIGEST_NONE;
    session->DataDigest = ISCSI_DIGEST_NONE;
    session->MaxRecvDataSegmentLength = DEFAULT_MAX_RECV_DATA_SEGMENT_LENGTH;
    session->MaxXmitDataSegmentLength = DEFAULT_MAX_RECV_DATA_SEGMENT_LENGTH;
    session->portal_group_tag = PORTAL_GROUP_TAG_UNKNOWN;
    session->update_address = &update_address;

    debugmsg(4, "sendtargets discovery process %p to %s:%d using "
	     "isid 0x%02x%02x%02x%02x%02x%02x",
	     process, config->address, config->port, session->isid[0],
	     session->isid[1], session->isid[2], session->isid[3],
	     session->isid[4], session->isid[5]);

    /* resolve the DiscoveryAddress to an IP address */
    while (!hostn) {
	hostn = gethostbyname(config->address);
	if (hostn) {
	    /* save the resolved address */
	    port = config->port;
	    ip_length = hostn->h_length;
	    memcpy(&ip_address, hostn->h_addr,
		   MIN(sizeof (ip_address), hostn->h_length));
	    /* FIXME: IPv6 */
	    debugmsg(4, "resolved %s to %u.%u.%u.%u\n", config->address,
		     ip_address[0], ip_address[1], ip_address[2],
		     ip_address[3]);
	} else {
	    errormsg("cannot resolve host name %s", config->address);
	    sleep(1);
	}
	if (iscsi_process_should_exit()) {
	    exit(0);
	}
    }

    sprintf(default_port, "%d", config->port);

    debugmsg(4,
	     "discovery timeouts: login %d, auth %d, active %d, idle %d, ping %d",
	     session->login_timeout, session->auth_timeout,
	     session->active_timeout, session->idle_timeout,
	     session->ping_timeout);

    /* if we have any incoming credentials, we insist on
     * authenticating the target or not logging in at all 
     */
    if (config->auth_options.username_in[0]
	|| config->auth_options.password_length_in) {
	session->bidirectional_auth = 1;

	/* sanity check the config */
	if ((config->auth_options.username[0] == '\0')
	    || (config->auth_options.password_length == 0)) {
	    logmsg(AS_ERROR,
		   "discovery process to %s:%d has incoming authentication "
		   "credentials but has no outgoing credentials configured\n",
		   config->address, config->port);
	    logmsg(AS_ERROR,
		   "discovery process %p to %s:%d exiting, bad configuration\n",
		   process, config->address, config->port);
	    exit(0);
	}
    } else {
	/* no or 1-way authentication */
	session->bidirectional_auth = 0;
    }

    /* copy in whatever credentials we have */
    strncpy(session->username, config->auth_options.username,
	    sizeof (session->username));
    session->username[sizeof (session->username) - 1] = '\0';
    if ((session->password_length = config->auth_options.password_length))
	memcpy(session->password, config->auth_options.password,
	       session->password_length);

    strncpy(session->username_in, config->auth_options.username_in,
	    sizeof (session->username_in));
    session->username_in[sizeof (session->username_in) - 1] = '\0';
    if ((session->password_length_in = config->auth_options.password_length_in))
	memcpy(session->password_in, config->auth_options.password_in,
	       session->password_length_in);

    if (session->password_length || session->password_length_in) {
	/* setup the auth buffers */
	session->auth_buffers[0].address = &session->auth_client_block;
	session->auth_buffers[0].length = sizeof (session->auth_client_block);

	session->auth_buffers[1].address = &session->auth_recv_string_block;
	session->auth_buffers[1].length =
	    sizeof (session->auth_recv_string_block);

	session->auth_buffers[2].address = &session->auth_send_string_block;
	session->auth_buffers[2].length =
	    sizeof (session->auth_send_string_block);

	session->auth_buffers[3].address = &session->auth_recv_binary_block;
	session->auth_buffers[3].length =
	    sizeof (session->auth_recv_binary_block);

	session->auth_buffers[4].address = &session->auth_send_binary_block;
	session->auth_buffers[4].length =
	    sizeof (session->auth_send_binary_block);

	session->num_auth_buffers = 5;
    } else {
	session->num_auth_buffers = 0;
    }

  set_address:
    /*
     * copy the saved address to the session,
     * undoing any temporary redirect
     */
    session->port = port;
    session->ip_length = ip_length;
    memcpy(session->ip_address, ip_address,
	   MIN(sizeof (session->ip_address), ip_length));

  reconnect:

    iscsi_disconnect(session);

    session->CmdSn = 1;
    session->itt = 1;
    session->portal_group_tag = PORTAL_GROUP_TAG_UNKNOWN;

    /* if we're violating the protocol anyway, there's no reason to be picky
     * about sending keys.
     */
    session->vendor_specific_keys = long_lived;

    /* slowly back off the frequency of login attempts */
    if (login_failures == 0)
	login_delay = 0;
    else if (login_failures < 10)
	login_delay = 1;	/* 10 seconds at 1 sec each */
    else if (login_failures < 20)
	login_delay = 2;	/* 20 seconds at 2 sec each */
    else if (login_failures < 26)
	login_delay = 5;	/* 30 seconds at 5 sec each */
    else if (login_failures < 34)
	login_delay = 15;	/* 60 seconds at 15 sec each */
    else
	login_delay = 60;	/* after 2 minutes, try once a minute */

    if (login_delay) {
	debugmsg(4, "sleeping for %d seconds before next login attempt",
		 login_delay);
	sleep(login_delay);
    }

    if (!iscsi_connect(session)) {
	/* FIXME: IPv6 */
	logmsg(AS_ERROR, "Connection to Discovery Address %u.%u.%u.%u failed",
	       session->ip_address[0], session->ip_address[1],
	       session->ip_address[2], session->ip_address[3]);

	login_failures++;
	/* If a temporary redirect sent us to something
	 * unreachable, we want to go back to the original
	 * IP address, so make sure we reset the session's IP.
	 */
	goto set_address;
    }
    logmsg(AS_NOTICE, "Connected to Discovery Address %u.%u.%u.%u",
	   session->ip_address[0], session->ip_address[1],
	   session->ip_address[2], session->ip_address[3]);

    debugmsg(4, "starting iSCSI login on fd %d", session->socket_fd);
    status_class = 0;
    status_detail = 0;
    switch (iscsi_login
	    (session, buffer_data(&sendtargets), unused_length(&sendtargets),
	     &status_class, &status_detail)) {
    case LOGIN_OK:
	break;

    case LOGIN_IO_ERROR:
    case LOGIN_WRONG_PORTAL_GROUP:
    case LOGIN_REDIRECTION_FAILED:
	/* try again */
	/* FIXME: IPv6 */
	logmsg(AS_NOTICE, "retrying discovery login to %u.%u.%u.%u",
	       session->ip_address[0], session->ip_address[1],
	       session->ip_address[2], session->ip_address[3]);
	iscsi_disconnect(session);
	login_failures++;
	goto set_address;

    default:
    case LOGIN_FAILED:
    case LOGIN_NEGOTIATION_FAILED:
    case LOGIN_AUTHENTICATION_FAILED:
    case LOGIN_VERSION_MISMATCH:
    case LOGIN_INVALID_PDU:
	/* FIXME: IPv6 */
	logmsg(AS_ERROR, "discovery login to %u.%u.%u.%u failed, giving up",
	       session->ip_address[0], session->ip_address[1],
	       session->ip_address[2], session->ip_address[3]);
	iscsi_disconnect(session);
	exit(0);
    }

    /* check the login status */
    switch (status_class) {
    case STATUS_CLASS_SUCCESS:
	/* FIXME: IPv6 */
	debugmsg(4, "discovery login success to %u.%u.%u.%u",
		 session->ip_address[0], session->ip_address[1],
		 session->ip_address[2], session->ip_address[3]);
	login_failures = 0;
	break;
    case STATUS_CLASS_REDIRECT:
	switch (status_detail) {
	    /* the session IP address was changed by the login
	     * library, so just try again with this portal
	     * config but the new address.
	     */
	case ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP:
	    /* FIXME: IPv6 */
	    logmsg(AS_NOTICE,
		   "discovery login temporarily redirected to %u.%u.%u.%u "
		   "port %d\n",
		   session->ip_address[0], session->ip_address[1],
		   session->ip_address[2], session->ip_address[3],
		   session->port);
	    goto reconnect;
	case ISCSI_LOGIN_STATUS_TGT_MOVED_PERM:
	    /* FIXME: IPv6 */
	    logmsg(AS_NOTICE,
		   "discovery login permanently redirected to %u.%u.%u.%u "
		   "port %d\n",
		   session->ip_address[0], session->ip_address[1],
		   session->ip_address[2], session->ip_address[3],
		   session->port);
	    /* make the new address permanent */
	    ip_length = session->ip_length;
	    memcpy(ip_address, session->ip_address,
		   MIN(sizeof (ip_address), session->ip_length));
	    port = session->port;

	    goto reconnect;
	default:
	    logmsg(AS_ERROR,
		   "discovery login rejected: redirection type 0x%x not "
		   "supported\n", status_detail);
	    goto set_address;
	}
	break;
    case STATUS_CLASS_INITIATOR_ERR:
	/* FIXME: IPv6 */
	logmsg(AS_ERROR,
	       "discovery login to %u.%u.%u.%u rejected: initiator "
	       "error (%02x/%02x), non-retryable, giving up",
	       session->ip_address[0], session->ip_address[1],
	       session->ip_address[2], session->ip_address[3], status_class,
	       status_detail);
	iscsi_disconnect(session);
	exit(0);
    case STATUS_CLASS_TARGET_ERR:
	/* FIXME: IPv6 */
	logmsg(AS_ERROR,
	       "discovery login to %u.%u.%u.%u rejected: target error (%02x/%02x)",
	       session->ip_address[0], session->ip_address[1],
	       session->ip_address[2], session->ip_address[3], status_class,
	       status_detail);
	iscsi_disconnect(session);
	login_failures++;
	goto reconnect;
    default:
	/* FIXME: IPv6 */
	logmsg(AS_ERROR,
	       "discovery login to %u.%u.%u.%u failed, response with unknown "
	       "status class 0x%x, detail 0x%x",
	       session->ip_address[0], session->ip_address[1],
	       session->ip_address[2], session->ip_address[3], status_class,
	       status_detail);
	iscsi_disconnect(session);
	login_failures++;
	goto reconnect;
    }

  rediscover:
    /* reinitialize */
    truncate_buffer(&sendtargets, 0);
    truncate_buffer(&info, 0);

    /* we're going to do a discovery regardless */
    clear_timer(&async_timer);

    /* ask for targets */
    if (!request_targets(session)) {
	goto reconnect;
    }
    active = 1;

    /* set timeouts */
    if (long_lived) {
	clear_timer(&connection_timer);
    } else {
	set_timer(&connection_timer,
		  session->active_timeout + session->ping_timeout);
    }

    /* prepare to poll */
    memset(&pfd, 0, sizeof (pfd));
    pfd.fd = session->socket_fd;
    pfd.events = POLLIN | POLLPRI;

    while (!iscsi_process_should_exit()) {

	/* check timers before blocking */
	if (timer_expired(&connection_timer)) {
	    if (long_lived || !lun_inventory_changed) {
		/* long-lived, or never finished the first
		 * exchange (might be long-lived) */
		clear_timer(&connection_timer);
		debugmsg(1, "discovery process %p reconnecting after "
			 "connection timeout", process);
		goto reconnect;
	    } else {
		logmsg(AS_NOTICE,
		       "discovery process %p session logout, connection timer "
		       "expired", process);
		iscsi_logout_and_disconnect(session);
		exit(0);
	    }
	}

	if (active) {
	    /* ignore the async timer, we're in the middle of a discovery */
	    timeout = msecs_until(&connection_timer);
	} else {
	    /* to avoid doing LUN probing repeatedly, try to
	     * merge multiple Async PDUs into one
	     * rediscovery by deferring discovery until a
	     * timeout expires.
	     */
	    if (timer_expired(&async_timer)) {
		debugmsg(4, "async timer expired, rediscovering");
		clear_timer(&async_timer);
		goto rediscover;
	    } else
		timeout =
		    soonest_msecs(NULL, &connection_timer, &async_timer);
	}

	/* block until we receive a PDU, a TCP FIN, a TCP RST, or a timeout */
	debugmsg(4, "discovery process %p to %s:%d polling fd %d, timeout "
		 "in %f seconds",
		 process, config->address, config->port, pfd.fd,
		 timeout / 1000.0);

	pfd.revents = 0;
	rc = poll(&pfd, 1, timeout);

	debugmsg(7, "discovery process %p to %s:%d returned from poll, rc %d",
		 process, config->address, config->port, rc);

	if (rc > 0) {
	    if (pfd.revents & (POLLIN | POLLPRI)) {
		/* put any PDU data into the sendtargets buffer for now */
		data = buffer_data(&sendtargets) + data_length(&sendtargets);
		end_of_data = data + unused_length(&sendtargets);

		timeout = msecs_until(&connection_timer);

		if (iscsi_recv_pdu
		    (session, pdu, sizeof (*pdu), data, end_of_data - data,
		     timeout)) {
		    /* process response, and record the ttt */
		    switch (pdu->opcode) {
		    case ISCSI_OP_TEXT_RSP:{
			    struct IscsiTextRspHdr *text_response =
				(struct IscsiTextRspHdr *) pdu;
			    int dlength = ntoh24(pdu->dlength);
			    int final =
				(text_response->flags & ISCSI_FLAG_FINAL)
				|| (text_response->ttt == RSVD_TASK_TAG);

			    debugmsg(4,
				     "discovery process %p received text "
				     "response, %d data bytes, ttt 0x%x, "
				     "final 0x%x",
				     process, dlength,
				     ntohl(text_response->ttt),
				     text_response->flags & ISCSI_FLAG_FINAL);

			    /* mark how much more data in
			     * the sendtargets buffer is now
			     * valid 
			     */
			    enlarge_data(&sendtargets, dlength);

			    /* process as much as we can right now */
			    process_sendtargets_response(&sendtargets, &info,
							 final,
							 lun_inventory_changed,
							 config->address,
							 default_port,
							 process->pipe_fd);

			    if (final) {
				/* SendTargets exchange is now complete */
				active = 0;
				
				/*
				 * From now on, after any reconnect, 
				 * assume LUNs may have changed
				 */
				lun_inventory_changed = 1;
			    } else {
				/* ask for more targets */
				if (!iterate_targets
				    (session, text_response->ttt))
				    goto reconnect;
			    }
			    break;
			}
		    case ISCSI_OP_ASYNC_EVENT:{
			    struct IscsiAsyncEvtHdr *async_hdr =
				(struct IscsiAsyncEvtHdr *) pdu;
			    int dlength = ntoh24(pdu->dlength);
			    short senselen;
			    char logbuf[128];
			    int i;

			    /*
			     * If we receive an async message stating
			     * the target wants to close the connection,
			     * then don't try to reconnect anymore.
			     * This is reasonable, so we don't log
			     * anything here.
			     */
			    if ((async_hdr->async_event ==
				  ASYNC_EVENT_REQUEST_LOGOUT)||
				 (async_hdr->async_event ==
				  ASYNC_EVENT_DROPPING_CONNECTION)||
				 (async_hdr->async_event ==
				  ASYNC_EVENT_DROPPING_ALL_CONNECTIONS)) {
				long_lived = 0;
				break;
			    }

			    /*
			     * Log information about async message.
			     */
			    logmsg(AS_NOTICE,
				   "Received Async Message from target, Event = %d, "
				   "Code = %d, Data Len = %d\n",
				   async_hdr->async_event,
				   async_hdr->async_vcode, dlength);

			    /*
			     * If there was data, print out the first 8 bytes
			     */
			    if (dlength > 0) {
				memset(logbuf, 0, sizeof(logbuf));
				for (i=0; i<8 && i<dlength; ++i) {
				    sprintf(logbuf+i*5, "0x%02x ", data[i]);
				}
				logmsg (AS_NOTICE, " Data[0]-[%d]: %s\n",
					i<dlength ? dlength-1 : i-1,
					logbuf);
			    }
					    

			    if (dlength > (sizeof (short))) {
				senselen = (ntohs(*(short *) (data)));

				debugmsg(1, " senselen = %d\n", senselen);
				if (dlength > senselen + 2) {
				    debugmsg(1, " received async event : %s\n",
					     data + 2 + senselen);
				}
			    }
			    long_lived = 1;
			    
			    /*
			     * Arrange for a rediscovery to occur in the near future.  We
			     * use a timer so that we merge multiple events that occur in
			     * rapid succession, and only rediscover once for each
			     * burst of Async events.
			     */
			    set_timer(&async_timer, 1);
			    if (active) {
				debugmsg(4,
					 "discovery process %p received Async "
					 "event while active", process);
			    } else {
				debugmsg(4,
					 "discovery process %p received Async "
					 "event while idle", process);
			    }
			    process_async_event_text(&sendtargets, &info,
						     &async_timer,
						     process->pipe_fd);
			    break;
			}
		    case ISCSI_OP_NOOP_IN:{
			    struct IscsiNopInHdr *nop =
				(struct IscsiNopInHdr *) pdu;

			    /*
			     * The iSCSI spec doesn't allow Nops on
			     * discovery sessions, but some targets
			     * use them anyway.  If we receive one, we
			     * can safely assume that the target
			     * supports long-lived discovery sessions
			     * (otherwise it wouldn't be sending nops
			     * to verify the connection is still
			     * working).
			     */
			    long_lived = 1;
			    debugmsg(4, "discovery process %p received Nop-in "
				     "with itt %u, ttt %u, dlength %u",
				     process, ntohl(nop->itt), ntohl(nop->ttt),
				     ntoh24(nop->dlength));

			    if (nop->ttt != RSVD_TASK_TAG) {
				/* reply to the Nop-in */
				if (!send_nop_reply
				    (session, nop, data,
				     session->active_timeout)) {
				    logmsg(AS_ERROR,
					   "discovery process %p failed to "
					   "send Nop reply, ttt %u, reconnecting",
					   process, ntohl(nop->ttt));
				    goto reconnect;
				}
			    }
			    break;
			}
		    case ISCSI_OP_REJECT_MSG:{
			    struct IscsiRejectRspHdr *reject =
				(struct IscsiRejectRspHdr *) pdu;
			    int dlength = ntoh24(pdu->dlength);
			    
			    logmsg(AS_ERROR, "reject, dlength=%d, data[0]=0x%x\n",
				   dlength, data[0]);
			    logmsg(AS_ERROR,
				   "Received a reject from the target with "
				   "reason code = 0x%x\n", reject->reason);
			    /*
			     * Just attempt to reconnect if we receive a reject
			     */
			    goto reconnect;
			    break;
			}
		    default:{
			    logmsg(AS_NOTICE,
				   "discovery process %p received unexpected "
				   "opcode 0x%x", process, pdu->opcode);
			    goto reconnect;
			}
		    }

		    /* reset timers after receiving a PDU */
		    if (long_lived) {
			clear_timer(&connection_timer);
		    } else {
			if (active)
			    set_timer(&connection_timer,
				      session->active_timeout);
			else
			    /* 3 minutes to try to go long-lived */
			    set_timer(&connection_timer, 3 * 60);
		    }
		} else {
		    if (long_lived) {
			debugmsg(1, "failed to recv a PDU response, reconnecting");
			goto reconnect;
		    } else {
			debugmsg(1, "failed to recv a PDU response, terminating");
			iscsi_disconnect(session);
			exit(0);
		    }
		}
	    }
	    if (pfd.revents & POLLHUP) {
		if (long_lived) {
		    logmsg(AS_NOTICE,
			   "discovery session to %s reconnecting after hangup",
			   session->TargetName);
		    goto reconnect;
		} else {
		    logmsg(AS_NOTICE,
			   "discovery session to %s terminating after hangup",
			   session->TargetName);
		    iscsi_disconnect(session);
		    exit(0);
		}
	    }

	    if (pfd.revents & POLLNVAL) {
		logmsg(AS_NOTICE, "discovery POLLNVAL");
		sleep(1);
		goto reconnect;
	    }

	    if (pfd.revents & POLLERR) {
		logmsg(AS_NOTICE, "discovery POLLERR");
		sleep(1);
		goto reconnect;
	    }
	} else if (rc < 0) {
	    if (errno == EINTR) {
		/* if we got SIGHUP, reconnect and rediscover */
		if (rediscover) {
		    rediscover = 0;
		    debugmsg(1, "rediscovery requested");
		    goto reconnect;
		}
	    } else {
		errormsg("poll error");
		sleep(5);
		exit(1);
	    }
	}
    }

    debugmsg(1, "discovery process %p to %s:%d exiting", process,
	     config->address, config->port);
}

static void
discovery_file_process(iscsi_discovery_process_t * process)
{
    iscsi_discovery_file_config_t *config = process->entry->config.file;
    struct sigaction action;
    int fd;
    int luns_changed = 0;
    string_buffer_t sendtargets;
    string_buffer_t info;

    process->pid = getpid();
    debugmsg(1,
	     "discovery file process %d, pid %d, file %s, username %s, password %s",
	     process->order, process->pid, config->filename,
	     config->auth_options.username, config->auth_options.password);

    if (!config->filename || !config->filename[0]) {
	logmsg(AS_ERROR, "no discovery filename specified");
	exit(0);
    }

    /* set SIGTERM, SIGINT, and SIGPIPE to kill the process */
    memset(&action, 0, sizeof (struct sigaction));
    action.sa_sigaction = NULL;
    action.sa_flags = 0;
    action.sa_handler = SIG_DFL;
    sigaction(SIGTERM, &action, NULL);
    sigaction(SIGINT, &action, NULL);
    sigaction(SIGPIPE, &action, NULL);

    /* set SIGHUP to loop again and report LUN INVENTORY CHANGED for testing */
    action.sa_handler = sighup_handler;
    sigaction(SIGHUP, &action, NULL);

    /* check if a signal arrived before we reset the handler */
    if (iscsi_process_should_exit()) {
	debugmsg(1, "discovery process %p exiting\n", process);
	exit(0);
    }

    /* allocate data buffers for SendTargets data and pipe info */
    init_string_buffer(&sendtargets, 32 * 1024);
    init_string_buffer(&info, 8 * 1024);

  repeat:
    truncate_buffer(&sendtargets, 0);
    truncate_buffer(&info, 0);

    if ((fd = open(config->filename, 0))) {
	char *data = buffer_data(&sendtargets);
	int rc;
	int final = 0;

	debugmsg(4, "opened discovery file %s", config->filename);

	do {
	    char *start = buffer_data(&sendtargets) + data_length(&sendtargets);
	    char *end = start + config->read_size;

	    /* don't overflow the buffer */
	    if (end > (start + unused_length(&sendtargets)))
		end = start + unused_length(&sendtargets);

	    data = start;

	    /* read a chunk of data */
	    do {
		rc = read(fd, data, end - data);
		if (rc > 0) {
		    data += rc;
		    enlarge_data(&sendtargets, rc);
		    debugmsg(7, "read %d bytes from %s", rc, config->filename);
		} else if (rc < 0) {
		    errormsg("error reading discovery file %s",
			     config->filename);
		    final = 1;
		    break;
		} else {
		    debugmsg(4, "read to the end of discovery file %s",
			     config->filename);
		    end = data;
		    final = 1;
		    break;
		}
	    } while (data < end);

	    /* convert all the whitespace to NULs */
	    for (data = start; data < end; data++) {
		if ((*data == '\n') || (*data == ' ') || (*data == '\t'))
		    *data = '\0';
	    }

	    /* process the data */
	    process_sendtargets_response(&sendtargets, &info, final,
					 luns_changed, config->address,
					 config->port, process->pipe_fd);

	} while (!final);
    } else {
	errormsg("couldn't open discovery file %s", config->filename);
    }

    while (config->continuous) {
	/* wait for a signal, then repeat */
	luns_changed = 1;
	debugmsg(7, "discovery file process waiting for signals");
	poll(NULL, 0, -1);
	if (rediscover) {
	    rediscover = 0;
	    goto repeat;
	}
    }

    exit(0);
}

#ifdef SLP_ENABLE

void
slp_discovery_process(iscsi_discovery_process_t * discovery)
{
    iscsi_slp_config_t *config = discovery->entry->config.slp;
    struct sigaction action;
    char *pl;
    unsigned short flag = 0;

    memset(&action, 0, sizeof (struct sigaction));
    action.sa_sigaction = NULL;
    action.sa_flags = 0;
    action.sa_handler = SIG_DFL;
    sigaction(SIGTERM, &action, NULL);
    sigaction(SIGINT, &action, NULL);
    sigaction(SIGPIPE, &action, NULL);

    action.sa_handler = sighup_handler;
    sigaction(SIGHUP, &action, NULL);

    if (iscsi_process_should_exit()) {
	debugmsg(1, "discovery process %p exiting\n", discovery);
	exit(0);
    }

    discovery->pid = getpid();

    pl = generate_predicate_list(discovery, &flag);

    while (1) {
	if (flag == SLP_MULTICAST_ENABLED) {
	    discovery->flag = SLP_MULTICAST_ENABLED;
	    slp_multicast_srv_query(discovery, pl, GENERIC_QUERY);
	}

	if (flag == SLP_UNICAST_ENABLED) {
	    discovery->flag = SLP_UNICAST_ENABLED;
	    slp_unicast_srv_query(discovery, pl, GENERIC_QUERY);
	}
	sleep(config->poll_interval);
    }

    exit(0);
}

#endif

void
discovery_process(iscsi_discovery_process_t * discovery)
{
    if (discovery->entry) {
	switch (discovery->entry->type) {
	case CONFIG_TYPE_DISCOVERY_FILE:
	    discovery_file_process(discovery);
	    break;
	case CONFIG_TYPE_SENDTARGETS:
	    sendtargets_discovery_process(discovery);
	    break;
	case CONFIG_TYPE_SLP:
#ifdef SLP_ENABLE
	    slp_discovery_process(discovery);
#else
	    logmsg(AS_ERROR, "this build does not support SLP");
	    exit(0);
#endif
	    break;
	default:
	    logmsg(AS_ERROR, "can't fork unexpected discovery type %d",
		   discovery->entry->type);
	    break;
	}
    }
}
