#ifndef ISCSI_CONFIG_H_
#define ISCSI_CONFIG_H_

/*
 * iSCSI driver for Linux
 * Copyright (C) 2002 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-config.h,v 1.7.2.1 2004/08/10 23:04:47 coughlan Exp $
 *
 * iSCSI configuration
 *
 */

#include "iscsiAuthClient.h"	/* for the username and password sizes */
#include "iscsi-common.h"	/* for ISCSI_MAX_LUN */

/* default iSCSI port number */
#define ISCSI_DEFAULT_PORT 3260

/* ISIDs now have a typed naming authority in them.  We use an OUI */
#define DRIVER_ISID_0  0x00
#define DRIVER_ISID_1  0x02
#define DRIVER_ISID_2  0x3D

/* default window size */
#define TCP_WINDOW_SIZE (256 * 1024)

/* the following structures store the options set in the config file.
 * a structure is defined for each logically-related group of options.
 * if you are adding a new option, first check if it should belong
 * to one of the existing groups.  If it does, add it.  If not, define
 * a new structure.
 */

/* all authentication-related options should be added to this structure.
 * this structure is per-session, and can be configured
 * by TargetName but not Subnet.
 */
typedef struct iscsi_auth_config {
    unsigned int authmethod;
    char username[iscsiAuthStringMaxLength];
    unsigned char password[iscsiAuthStringMaxLength];
    unsigned int password_length;
    char username_in[iscsiAuthStringMaxLength];
    unsigned char password_in[iscsiAuthStringMaxLength];
    unsigned int password_length_in;
} iscsi_auth_config_t;

/* all per-connection timeouts go in this structure.
 * this structure is per-portal, and can be configured
 * both by TargetName and Subnet.
 */
typedef struct iscsi_connection_timeout_config {
    int login_timeout;
    int auth_timeout;
    int active_timeout;
    int idle_timeout;
    int ping_timeout;
} iscsi_connection_timeout_config_t;

/* all per-connection timeouts go in this structure.
 * this structure is per-session, and can be configured
 * by TargetName but not by Subnet.
 */
typedef struct iscsi_session_timeout_config {
    int replacement_timeout;
    unsigned int disk_command_timeout;
#ifdef SOLARIS
    int conn_fail_timeout;
#endif
} iscsi_session_timeout_config_t;

/* all error handling timeouts go in this structure.
 * this structure is per-portal, and can be configured
 * both by TargetName and Subnet.
 */
typedef struct iscsi_error_timeout_config {
    int abort_timeout;
    int reset_timeout;
} iscsi_error_timeout_config_t;

/* all TCP options go in this structure.
 * this structure is per-portal, and can be configured
 * both by TargetName and Subnet.
 */
typedef struct iscsi_tcp_config {
    int window_size;
    int type_of_service;	/* try to set IP TOS bits */
} iscsi_tcp_config_t;

/* all iSCSI operational params go in this structure.
 * this structure is per-portal, and can be configured
 * both by TargetName and Subnet.
 */
typedef struct iscsi_operational_config {
    int protocol;
    int InitialR2T;
    int ImmediateData;
    int MaxRecvDataSegmentLength;
    int FirstBurstLength;
    int MaxBurstLength;
    int DefaultTime2Wait;
    int DefaultTime2Retain;
    int HeaderDigest;
    int DataDigest;
} iscsi_operational_config_t;

#define CONFIG_DIGEST_NEVER  0
#define CONFIG_DIGEST_ALWAYS 1
#define CONFIG_DIGEST_PREFER_ON 2
#define CONFIG_DIGEST_PREFER_OFF 3

/* all multipath options go in this structure.
 * this structure is per-session, and can be configured
 * by TargetName but not by Subnet.
 */
typedef struct iscsi_multipath_config {
    int type;
} iscsi_multipath_config_t;

typedef struct iscsi_nics_for_mp {
    struct iscsi_nics_for_mp *next;
    char *address;
    int iscsi_bus;
} iscsi_nics_mp_t;

#define NO_MULTIPATH          0
#define MULTIPATH_NO          NO_MULTIPATH
#define MULTIPATH_PORTAL      1
#define MULTIPATH_PORTALGROUP 2
#define MULTIPATH_MANUAL      3

/* the following structures represent the contents of the config file */

typedef struct iscsi_targetname_config {
    char *TargetName;
    int enabled;
    int portal_failover;
    char *preferred_portal;
    char *preferred_subnet;
    uint32_t preferred_subnet_mask;
    uint32_t luns_allowed[(ISCSI_MAX_LUN + 31) / 32];
    iscsi_auth_config_t auth_options;
    iscsi_multipath_config_t multipath_options;
    iscsi_tcp_config_t tcp_options;
    iscsi_connection_timeout_config_t connection_timeout_options;
    iscsi_session_timeout_config_t session_timeout_options;
    iscsi_error_timeout_config_t error_timeout_options;
    iscsi_operational_config_t iscsi_options;
} iscsi_targetname_config_t;

typedef struct iscsi_subnet_config {
    char *address;
    int ip_length;
    char ip_address[16];
    int port;
    uint32_t subnet_mask;	/* IPv4 subnet mask */
    iscsi_connection_timeout_config_t connection_timeout_options;
    iscsi_error_timeout_config_t error_timeout_options;
    iscsi_tcp_config_t tcp_options;
} iscsi_subnet_config_t;

typedef struct iscsi_sendtargets_config {
    char *address;
    int port;
    int continuous;
    int send_async_text;
    iscsi_auth_config_t auth_options;
    iscsi_connection_timeout_config_t connection_timeout_options;
} iscsi_sendtargets_config_t;

typedef struct iscsi_slp_config {
    char *address;		/* for unicast */
    int port;			/* for unicast */
    char *scopes;
    char *interfaces;		/* for multicast, list of
				 * interfaces names, "all",
				 * or "none" 
				 */
    int poll_interval;
    iscsi_auth_config_t auth_options;
} iscsi_slp_config_t;

typedef struct iscsi_discovery_file_config {
    char *filename;
    int read_size;
    char *address;
    char *port;
    int continuous;
    iscsi_auth_config_t auth_options;
} iscsi_discovery_file_config_t;

/* config defaults */
typedef struct iscsi_config_defaults {
    /* discovery defaults */
    int continuous_sendtargets;	/* if non-zero, default to
				 * keeping iSCSI discovery
				 * sessions open 
				 */
    int send_async_text;	/* if non-zero, target will
				 * send vendor specific
				 * async events. 
				 */
    int slp_multicast;		/* if non-zero, default to
				 * attempting SLP multicast
				 * discovery of all targets
				 * this initiator can
				 * access 
				 */
    char *slp_scopes;
    int slp_poll_interval;

    /* global default options, used if no options are in the config,
     * or if the user sets global defaults in the config file.
     */
    int enabled;
    uint32_t luns_allowed[(ISCSI_MAX_LUN + 31) / 32];
    iscsi_multipath_config_t multipath_options;
    iscsi_auth_config_t auth_options;
    iscsi_connection_timeout_config_t connection_timeout_options;
    iscsi_error_timeout_config_t error_timeout_options;
    iscsi_session_timeout_config_t session_timeout_options;
    iscsi_tcp_config_t tcp_options;
    iscsi_operational_config_t iscsi_options;
    int portal_failover;
    char *preferred_portal;
    char *preferred_subnet;
    uint32_t preferred_subnet_mask;
} iscsi_config_defaults_t;

typedef struct iscsi_config_entry {
    struct iscsi_config_entry *prev;
    struct iscsi_config_entry *next;
    int type;
    int line_number;
    union {
	iscsi_sendtargets_config_t *sendtargets;
	iscsi_slp_config_t *slp;
	iscsi_discovery_file_config_t *file;
	iscsi_targetname_config_t *targetname;
	iscsi_subnet_config_t *subnet;
    } config;
} iscsi_config_entry_t;

#define CONFIG_TYPE_UNKNOWN        0
#define CONFIG_TYPE_SENDTARGETS    1
#define CONFIG_TYPE_SLP            2
#define CONFIG_TYPE_DISCOVERY_FILE 3
#define CONFIG_TYPE_TARGETNAME     4
#define CONFIG_TYPE_SUBNET         5
#define CONFIG_TYPE_ADDRESS        6

/* collect all iscsi config file info together */
typedef struct iscsi_config {
    iscsi_config_defaults_t defaults;
    iscsi_config_entry_t *head;
    iscsi_config_entry_t *tail;
} iscsi_config_t;

/* the following structures represent the run-time configuration,
 * computed based on the discovery info and the structures representing
 * the config file contents.
 */

/* discovery produces portal descriptors.
 * this arguably belongs in a different header file,
 * but is here since everything references them
 * via portal_configs
 */
typedef struct iscsi_portal_descriptor {
    struct iscsi_portal_descriptor *next;
    char *address;		/* text string */
    char ip[16];		/* binary IP */
    int ip_length;
    int port;
    int tag;
} iscsi_portal_descriptor_t;

#define PORTAL_GROUP_TAG_UNKNOWN -1

/* structures dynamically created as the main daemon
 * collects discovery info and processes config info 
*/

typedef struct iscsi_portal_config {
    struct iscsi_portal_config *next;

    iscsi_portal_descriptor_t *descriptor;	/* the target_config's
						 * portal descriptor 
						 */

    /* and the config options to use when the connection uses this portal */
    iscsi_connection_timeout_config_t connection_timeout_options;
    iscsi_session_timeout_config_t session_timeout_options;
    iscsi_error_timeout_config_t error_timeout_options;
    iscsi_tcp_config_t tcp_options;
    iscsi_operational_config_t iscsi_options;
} iscsi_portal_config_t;

typedef struct iscsi_portal_config_list {
    struct iscsi_portal_config *head;
    struct iscsi_portal_config *tail;
} iscsi_portal_config_list_t;

/* config code doesn't need to care about what's in the session process struct */
struct iscsi_session_process;

struct iscsi_target_config;

/* a normal iSCSI session */
typedef struct iscsi_session_config {
    struct iscsi_session_config *next;
    struct iscsi_session_process *process;
    struct iscsi_target_config *target;
    iscsi_portal_config_list_t portals;
    char *nic_ip_address;
    int iscsi_bus;
    int target_id;
    int binding_index;
    unsigned char isid[6];
    int path_number;
} iscsi_session_config_t;

struct iscsi_discovery_process;

typedef struct iscsi_target_config {
    char *TargetName;		/* typically shared with
				 * some other structure,
				 * which manages the
				 * lifetime 
				 */

    /* options that only make sense for the target as a whole */
    int enabled;
    int portal_failover;
    char *preferred_portal;
    char *preferred_subnet;
    uint32_t preferred_subnet_mask;
    uint32_t luns_allowed[(ISCSI_MAX_LUN + 31) / 32];
    iscsi_multipath_config_t multipath_options;
    iscsi_auth_config_t auth_options;

    iscsi_session_config_t *sessions;

} iscsi_target_config_t;

/* exported functions */
extern char *get_iscsi_initiatorname(char *pathname);

extern int update_iscsi_config(const char *pathname, iscsi_config_t * config);

extern int add_config_entry(iscsi_config_t * config,
			    iscsi_config_entry_t * entry);
extern int remove_config_entry(iscsi_config_t * config,
			       iscsi_config_entry_t * entry);
extern void free_config_entry(iscsi_config_entry_t * entry);

extern iscsi_target_config_t *create_target_config(char *name,
						   iscsi_portal_descriptor_t *
						   portals,
						   iscsi_config_t * config,
						   iscsi_auth_config_t *
						   auth_options);

extern void free_target_config(iscsi_target_config_t * config);

extern void create_session_configs(iscsi_target_config_t * target,
				   iscsi_portal_descriptor_t * portals,
				   iscsi_config_t * config);

extern void free_session_config(iscsi_session_config_t * config);

extern void free_portal_descriptors(iscsi_portal_descriptor_t * portals);

/* comparisons */
extern int same_portal_descriptor(iscsi_portal_descriptor_t * p1,
				  iscsi_portal_descriptor_t * p2);
extern int same_portal_descriptors(iscsi_portal_descriptor_t * portals1,
				   iscsi_portal_descriptor_t * portals2);
extern int same_portal_config(iscsi_portal_config_t * p1,
			      iscsi_portal_config_t * p2);
extern int same_portal_configs(iscsi_portal_config_t * portals1,
			       iscsi_portal_config_t * portals2);
extern int same_session_config(iscsi_session_config_t * s1,
			       iscsi_session_config_t * s2);
extern int same_target_config(iscsi_target_config_t * t1,
			      iscsi_target_config_t * t2);

#endif
