/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-device.c,v 1.4.2.1 2004/08/10 23:04:49 coughlan Exp $
 *
 * iscsi-device.c
 *
 *     Determine if a Linux SCSI device is an iSCSI device.
 *
 * May 2001, sferris
 *
 * Added the ability to get the page80 LUN serial # or page83 LUN Device
 * ID for the device in question (Generically referred to as WWID)
 * AJ Lewis <alewis@redhat.com> (6/17/2004)
 * Copyright (C) 2004 Red Hat, Inc
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

typedef uint8_t u8;
#include <scsi/scsi.h>
#include <scsi/scsi_ioctl.h>

#include "scsi-info.h"

/* if the device name on the command line exists and is an iSCSI device,
 * then output a description of the iSCSI target to stdout, and exit 0.
 * if the device name might be an iSCSI device, exit 1.
 * if the device name definately is not an iSCSI device, exit 2.
 */

#define ISCSI_PROC_DIR "/proc/scsi/iscsi"

int
main(int argc, char *argv[])
{
    char devname[PATH_MAX + 1];
    char buffer[4097];
    char *device = argv[1];
    int iscsi_channel = 0;
    int iscsi_id = 0;
    int iscsi_lun = 0;
    FILE *proc_hba;
    char *line;
    char *link;
    struct scsi_info info;

    memset(&info, 0, sizeof(info));

    if (argc > 2) {
	fprintf(stderr,
		"iscsi-device: error, too many arguments, only 1 device name "
		"is allowed.\n");
    }
    if (argc != 2) {
	fprintf(stderr, "\nUsage: iscsi-device devicename\n\n");
	fprintf(stderr,
		"If the device name refers to an iSCSI device, iscsi-device\n");
	fprintf(stderr,
		"  returns 0 after writing the following information to stdout:\n");
	fprintf(stderr,
		"\tthe resolved devicename\n\tdevlabel link (if it exists)\n\tiscsi host number\n\tscsi channel "
		"number\n\ttarget id number\n\tLUN number\n");
	fprintf(stderr,
		"\ttarget ip address\n\ttarget port\n\tiscsi target id\n\tiSCSI "
		"target name\n\tLUN WWID (if it exists)\n\n");
	fprintf(stderr,
		"If the device name does may be an iSCSI device not yet ready, "
		"iscsi-device\n");
	fprintf(stderr,
		"  returns error code 1, and may write a message to stderr.\n\n\n");
	fprintf(stderr,
		"If the device name definately does not refer to an iSCSI device, "
		"iscsi-device\n");
	fprintf(stderr,
		"  returns an error code greater than 1, and may write a message "
		"to stderr.\n\n\n");
	exit(2);
    }
    info.path = strdup(argv[1]);
    /* get the (host,channel,id,lun) quadruple for this device, if possible */
    get_scsi_info(&info);
    link = get_devlabel_link(&info);
    /* see if the device node's host (HBA number) is an iSCSI HBA */
    sprintf(devname, "%s/%d", ISCSI_PROC_DIR, info.host);
    if ((proc_hba = fopen(devname, "r"))) {
	/* read every line, ignoring comments and blank lines */
	while ((line = fgets(buffer, sizeof (buffer), proc_hba))) {
            line[strlen(line)-1] = '\0';
	    while (isspace(*line))
		line++;
	    if (*line == '#' || *line == '\0')
		continue;

	    if (sscanf(line, "%d %d %d", &iscsi_channel, &iscsi_id, &iscsi_lun)
		== 3) {
		/* For now we assume there is only one host,
		 * so that the bus number maps directly to
		 * the channel number.  This breaks if we
		 * ever have multiple HBAs.
		 */
		if (info.channel == iscsi_channel &&
		    info.id == iscsi_id && info.lun == iscsi_lun) {
		    /* found it. Print the device and /proc line */
		    printf("%s%s%s%s: %s", info.path, 
                           link ? " (" : "",
                           link ? link : "",
                           link ? ")" : "",
                           line);
		    /* print wwid */
		    if (info.page83_type3)
			printf(" S83.3:%s\n", info.page83_type3);
		    else if (info.page83_type2)
			printf(" S83.2:%s\n", info.page83_type2);
		    else if (info.page83_type1)
			printf(" S83.1:%s\n", info.page83_type1);
		    else if (info.page80)
			printf(" page 80: %s\n", info.page80);
		    else if (info.page83_type0)
			printf(" S83.3:%s\n", info.page83_type0);
		    exit(0);
		}

	    }
	}
	fclose(proc_hba);

	/* if /proc/scsi/iscsi has a file for the device node's host,
	 * then we know it's an iSCSI device, even if we can't get any
	 * additional info about it.
	 */
	exit(0);
    }

    /* didn't find a match */
    fprintf(stderr, "iscsi-device: %s is not an iSCSI device\n", device);
    exit(2);
}
