/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * iscsi-ls.c - User level utility to display iscsi devices.
 * $Id: iscsi-ls.c,v 1.3.4.6 2005/06/14 19:26:28 vortechs Exp $
 */

/* Added ability to get page80 and page83 scsi info on a per LUN
 * basis. This info shows up with the -l flag.
 * ajl <alewis@redhat.com> (6/17/2004)
 * Copyright (C) 2004 Red Hat, Inc
 */

#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include <dirent.h>
#include "iscsi-version.h"
#include "iscsi-common.h"
#include "iscsi-ioctl.h"
#include "scsi-info.h"

#define BUS_SPECIFIC 	1
#define TARGET_SPECIFIC 2
#define TEXT_KEYS       4
#define LUN_INFO        8

const char *bindings_file = "/var/lib/iscsi/bindings";

void
print_usage(void)
{
    printf("\nUsage: iscsi-ls [-b busId] [-t targetId] [-c] [-l]\n");
    printf("-b : Details regarding the specific bus\n");
    printf("-t : Details regarding the specific target\n");
    printf("-c : iSCSI target configuration parameters\n");
    printf("-l : Lun details available on the target\n");
    return;
}

static int
device_matches(struct target_info ti, int lun_id, struct scsi_info si)
{
    if(ti.target_id != si.id)
        return 0;
    if(ti.channel != si.channel)
        return 0;
    if(lun_id != si.lun)
        return 0;
    if(ti.host_no != si.host)
        return 0;
    return 1;
}


/* Read in the /oroc/partitions entry making sure it hasn't changed
 * while we're reading it. */
static char *
_create_procbuf(unsigned int *pbsize, unsigned int *count)
{
    int procfd = -1;
    char *pb = NULL;
    int tmpcount = 0;
    int tmpsize = 0;

    if((procfd = open("/proc/partitions", O_RDONLY)) == -1){
        fprintf(stderr, "Could not open /proc/partitions");
        goto end;
    }

    if(!(pb = (char *) malloc(sizeof(char) * (*pbsize)))) {
        fprintf(stderr, "Out of memory allocating buffer (%u)"
                " for /proc/partitions",
                *pbsize);
        goto end;
    }

    /* Read and realloc until we read less than we asked for */
    do {
        memset(pb, 0, sizeof(char) * (*pbsize));
        tmpsize = *pbsize;
        *count = 0;
        lseek(procfd, 0, SEEK_SET);
        while((tmpcount = read(procfd, pb+*count, tmpsize)) > 0) {
            tmpsize -= tmpcount;
            *count += tmpcount;
            if(tmpsize <= 0)
                break;
        }

        if(tmpcount < 0 && errno != EINTR) {
            free(pb);
            pb = NULL;
            goto end;
        }
        if(*count < *pbsize) {
            /* check to see if there was more - if there was, reread
             * the whole thing in case we were interrupted by a signal
             * - we want to get all of /proc/partitions in a single
             * read... */
            if((tmpcount = read(procfd, pb+*count, 1)) > 0) {
                continue;
            }
            else
                break;
        }

        *pbsize *= 2;
        if(!(pb = (char *) realloc(pb, sizeof(char) * (*pbsize)))) {
            fprintf(stderr, "Out of memory reallocating buffer (%u)"
                    "for /proc/partitions",
                    *pbsize);
            goto end;
        }

    } while(*pbsize < SSIZE_MAX);

  end:
    close(procfd);
    return pb;
}

static int
_get_devs_from_proc(struct scsi_info *info_head)
{
    char *procbuf = NULL, *tmpproc = NULL, *curproc = NULL;
    int size = 4096, count = 0, i;
    char maj[3], min[3], blocks[30], name[64],
        tmp_str[PATH_MAX], proc_line[PATH_MAX];
    struct scsi_info *info;

    if(!(procbuf = _create_procbuf(&size, &count)))
	    return 0;

    curproc = procbuf;

    /* Skip the first two lines of /proc/partitions */
    for(i = 0; i < 2; i++) {
        /* FIXME: should i use strchr? */
        tmpproc = memchr(curproc, '\n', PATH_MAX);
        curproc = tmpproc+1;
    }

    while((tmpproc = memchr(curproc, '\n', PATH_MAX)) &&
          (tmpproc - procbuf <= count)) {
        strncpy(proc_line, curproc, tmpproc-curproc);

        sscanf(proc_line,"%3s %3s %30s %64s", maj, min, blocks, name);

        /* FIXME: If there are > 26 scsi devices, we won't get those
           that have 2 letters in their names */
        if(!isalpha(name[0]) || strncmp(name, "sd", 2) || strlen(name) > 4 || (name[3] != '\0' && !isalpha(name[3]))){
            memset(name, 0, sizeof(name));
            curproc = tmpproc + 1;
            memset(proc_line, 0, sizeof(proc_line));
            continue;
        }

        memset(tmp_str, 0, sizeof(tmp_str));

        snprintf(tmp_str, sizeof(tmp_str), "/dev/%s", name);
        tmp_str[PATH_MAX-1] = '\0';

        /* FIXME: list processing fxns */
        info = malloc(sizeof(*info));
        memset(info, 0, sizeof(*info));
        info->n = info_head->n;
        info_head->n = info;
        info->path = strdup(tmp_str);
        info->id = -1;

        curproc = tmpproc + 1;
        memset(proc_line, 0, sizeof(proc_line));
    }
    free(procbuf);
    return 1;
}

static int
get_current_scsi_info(struct scsi_info *info_head)
{

    struct scsi_info *n;

    /* FIXME: Eventually we can add other ways to find the device
     * paths (udev, sysfs), with the preferred methods first and
     * failing back to the alternatives  */
    _get_devs_from_proc(info_head);

    list_iterate(n, info_head)
        get_scsi_info(n);

    return 1;
}

int
print_lun_info(struct target_info target_details, struct scsi_info *info_head)
{
    FILE *fp;
    char *rdbufp, rdbuf[1024];
    struct scsi_info *n;
    int lun_id = -1;
    char *link;

    fp = fopen("/proc/scsi/scsi", "r");
    if (!fp) {
	perror("\nCannot open /proc/scsi/scsi\n");
	return -1;
    }

    printf("\nDEVICE DETAILS :\n");
    printf("-------------- \n");
    do {
	int target_id;
	rdbufp = fgets(rdbuf, sizeof (rdbuf), fp);
	if (rdbufp && (*rdbufp != '#')) {
	    while (isspace(*rdbufp)) {
		rdbufp += 1;
	    }
	    if (!strncasecmp(rdbufp, "Attached", 8))
		continue;
	    if (!strncasecmp(rdbufp, "Host: scsi", 10)) {
		int channel_id;
		int host_id;
		char *cp = &rdbufp[10];
		sscanf(cp, "%d", &host_id);
		if (host_id != target_details.host_no) {
		    continue;
		}

		cp = &rdbufp[22];
		sscanf(cp, "%d", &channel_id);
		if (channel_id != target_details.channel) {
		    continue;
		}

		cp = &rdbufp[28];
		if (sscanf(cp, "%d", &target_id)) {
		    if (target_details.target_id == target_id) {
			cp = &rdbufp[36];
			sscanf(cp, "%d", &lun_id);
			printf("LUN ID : " "%d\n", lun_id);
			rdbufp = fgets(rdbuf, sizeof (rdbuf), fp);
			printf("%s", rdbufp);
			rdbufp = fgets(rdbuf, sizeof (rdbuf), fp);
			printf("%s", rdbufp);

                        list_iterate(n, info_head) {
                            if(device_matches(target_details, lun_id, *n)) {
                                if(n->page83_type3)
                                    printf("  page83 type3: %s\n",
                                           n->page83_type3);
                                if(n->page83_type2)
                                    printf("  page83 type2: %s\n",
                                           n->page83_type2);
                                if(n->page83_type1)
                                    printf("  page83 type1: %s\n",
                                           n->page83_type1);
                                if(n->page80)
                                    printf("  page80: %s\n", n->page80);
                                if(n->page83_type0)
                                    printf("  page83 type0: %s\n",
                                           n->page83_type0);

                                printf("  Device: %s\n", n->path);
                                if((link = get_devlabel_link(n))) {
                                    printf("  Devlabel: %s\n", link);
                                    free(link);
                                }
                            }
                        }
                    }
                }
            }
	}
    } while (rdbufp);
    fclose(fp);
    return 0;
}

int
main(int argc, char **argv)
{
    int i, j, k;
    int c, options = 0;
    int fd, ctlfd;
    int target_id = 0, bus_id = 0, id, max_busid = 0, max_targetid = 0;
    int result, rc;
    struct target_info target_details;
    FILE *fp;
    char *rdbufp, *cp, rdbuf[256], version[256];
    struct dirent *dirent;
    DIR *dirp;
    char proc_file[512], dir[128];
    struct scsi_info sinfo_head, *n;
    list_init(&sinfo_head);

    while ((c = getopt(argc, argv, "lct:b:")) >= 0) {
	switch (c) {
	case 'b':
	    bus_id = atoi(optarg);
	    options |= BUS_SPECIFIC;
	    break;
	case 't':
	    target_id = atoi(optarg);
	    if (target_id < 0 || target_id >= ISCSI_MAX_TARGETS) {
		printf("Target ids should be between 0-%d\n",
		       ISCSI_MAX_TARGETS - 1);
		return -1;
	    }
	    options |= TARGET_SPECIFIC;
	    break;
	case 'c':
	    options |= TEXT_KEYS;
	    break;
	case 'l':
	    options |= LUN_INFO;
	    break;
	default:
	    print_usage();
	    return -1;
	}
    }

    ctlfd = open("/dev/iscsictl", O_RDONLY);

    if (ctlfd < 0) {
	printf("\n\n#######################################"
	       "########################################\n");
	printf("\niSCSI driver is not loaded\n");
	printf("\n\n#######################################"
	       "########################################\n\n\n");
	return -1;
    }
#if defined(DRIVER_INTERNAL_VERSION) && (DRIVER_INTERNAL_VERSION > 0)
    sprintf(version, "SFNet iSCSI Driver Version ... %d.%d.%d.%d%s (%s )",
	    DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION,
	    DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
	    DRIVER_EXTRAVERSION, ISCSI_DATE);
#else
    sprintf(version, "SFNet iSCSI Driver Version ... %d.%d.%d%s (%s )",
	    DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION,
	    DRIVER_PATCH_VERSION, DRIVER_EXTRAVERSION, ISCSI_DATE);
#endif

    printf("**************************************"
	   "*****************************************\n");
    printf("\t%s\n", version);
    printf("**************************************"
	   "*****************************************\n");

    /* Finding the maximum target id and bus id in the system */
    sprintf(dir, "%s", "/proc/scsi/iscsi");
    dirp = opendir(dir);
    if (dirp == NULL) {
	fprintf(stderr, "Directory %s doesn't exist\n", dir);
	exit(1);
    }
    while ((dirent = readdir(dirp)) != NULL) {
	if (strcmp(dirent->d_name, ".") && strcmp(dirent->d_name, "..")) {
	    sprintf(proc_file, "/proc/scsi/iscsi/");
	    strcat(proc_file, dirent->d_name);
	    /* Finding the maximum target id and bus id in the system */
	    fd = open(proc_file, O_RDWR);
	    if (fd < 0) {
		fprintf(stderr, "Can't open %s - %s\n", proc_file,
			strerror(errno));
		return -1;
	    }
	    fp = fdopen(fd, "r+");
	    if (fp == NULL) {
		fprintf(stderr, "Can't open %s - %s\n", bindings_file,
			strerror(errno));
		close(fd);
		return -1;
	    }
	    while (((rdbufp = fgets(rdbuf, sizeof (rdbuf), fp)) != NULL)) {
		cp = strtok(rdbufp, " \t\n");
		if ((cp == NULL) || ((result = sscanf(cp, "%d", &id)) != 1)) {
		    continue;
		}
		max_busid = MAX(max_busid, id);
		cp = strtok(NULL, " \t\n");
		if ((cp == NULL) || ((result = sscanf(cp, "%d", &id)) != 1)) {
		    continue;
		}
		max_targetid = MAX(max_targetid, id);
	    }
	}
    }

    if (options & LUN_INFO) {
        /* Cache up scsi channel, lun, id, host info */

        get_current_scsi_info(&sinfo_head);
    }
    for (i = 0; i <= max_targetid; i++) {
	for (k = 0; k <= max_busid; k++) {
	    struct timeval tv;
	    struct iscsi_portal_list iscsi_ls_portal_list;
	    /* Get details on this target_id */
	    memset(&target_details, 0, sizeof (target_details));
	    if (options & BUS_SPECIFIC) {
		target_details.channel = bus_id;
	    } else {
		target_details.channel = k;
	    }
	    if (options & TARGET_SPECIFIC) {
		target_details.target_id = target_id;
	    } else {
		target_details.target_id = i;
	    }
	    rc = ioctl(ctlfd, ISCSI_LS_TARGET_INFO, &target_details);
	    if (rc) {
		fprintf(stderr, "The iscsi driver is inoperative\n");
		return -1;
	    }

	    if (options & BUS_SPECIFIC) {
		if (bus_id != target_details.channel) {
		    return -1;
		}
	    }

	    /* This target_id does not have a corresponding session.
	     * proceed with next.
	     */
	    if (strcmp(target_details.target_name, "") == 0) {
		if (options & TARGET_SPECIFIC) {
		    return -1;
		} else {
		    continue;
		}
	    }

	    printf("TARGET NAME             : %s\n",
		   target_details.target_name);
	    printf("TARGET ALIAS            : %s\n",
		   target_details.target_alias);
	    printf("HOST NO                 : %d \n", target_details.host_no);
	    printf("BUS NO                  : %d \n", target_details.channel);
	    printf("TARGET ID               : %d \n", target_details.target_id);
	    printf("TARGET ADDRESS          : %u.%u.%u.%u:%d\n",
		   target_details.session_data.addr[0],
		   target_details.session_data.addr[1],
		   target_details.session_data.addr[2],
		   target_details.session_data.addr[3],
		   target_details.session_data.port);
	    if (!target_details.session_data.session_alive) {
		if (!target_details.session_data.ever_established) {
		    printf("SESSION STATUS          : " "NOT ESTABLISHED\n");
		} else {
		    gettimeofday(&tv, NULL);
		    tv.tv_sec -= target_details.session_data.session_drop_time;
		    printf("SESSION STATUS          : "
			   "DROPPED AT %s", ctime((time_t *) & (tv.tv_sec)));
		}
	    } else {
		gettimeofday(&tv, NULL);
		tv.tv_sec -= target_details.session_data.establishment_time;
		printf("SESSION STATUS          : "
		       "ESTABLISHED AT %s", ctime((time_t *) & (tv.tv_sec)));
	    }
	    printf("NO. OF PORTALS          : %d \n",
		   target_details.num_portals);
	    memset(&iscsi_ls_portal_list, 0, sizeof (iscsi_ls_portal_list));
	    iscsi_ls_portal_list.target_id = target_details.target_id;
	    iscsi_ls_portal_list.bus_id = target_details.channel;
	    iscsi_ls_portal_list.portals =
		(struct iscsi_portal_info *) malloc(target_details.
						    num_portals *
						    sizeof (struct
							    iscsi_portal_info));
	    if (!iscsi_ls_portal_list.portals) {
		fprintf(stderr, "Failed to allocate memory\n");
		return -1;
	    }
	    rc = ioctl(ctlfd, ISCSI_LS_PORTAL_INFO, &iscsi_ls_portal_list);
	    if (rc) {
		fprintf(stderr, "The iscsi driver is inoperative\n");
		free(iscsi_ls_portal_list.portals);
		return -1;
	    }

	    for (j = 0; j < target_details.num_portals; j++) {
		printf("PORTAL ADDRESS %d        : %u.%u.%u.%u:"
		       "%d,%d\n",
		       (j + 1),
		       iscsi_ls_portal_list.portals[j].
		       ip_address[0],
		       iscsi_ls_portal_list.portals[j].
		       ip_address[1],
		       iscsi_ls_portal_list.portals[j].
		       ip_address[2],
		       iscsi_ls_portal_list.portals[j].
		       ip_address[3],
		       iscsi_ls_portal_list.portals[j].port,
		       iscsi_ls_portal_list.portals[j].tag);
	    }
	    free(iscsi_ls_portal_list.portals);
	    printf("SESSION ID              : ISID %02x%02x%02x%02x"
		   "%02x%02x TSIH %02x\n",
		   target_details.session_data.isid[0],
		   target_details.session_data.isid[1],
		   target_details.session_data.isid[2],
		   target_details.session_data.isid[3],
		   target_details.session_data.isid[4],
		   target_details.session_data.isid[5],
		   target_details.session_data.tsih);

	    if (options & TEXT_KEYS) {
		printf("\nTARGET CONFIGURATION KEYS :\n");
		printf("------------------------- \n");
		printf("INITIAL R2T             : %s\n",
		       (target_details.session_data.InitialR2T) ? "Yes" : "No");
		printf("IMMEDIATE DATA          : %s\n",
		       (target_details.session_data.
			ImmediateData) ? "Yes" : "No");
		printf("HEADERDIGEST            : %s\n",
		       (target_details.session_data.
			HeaderDigest) ? "CRC32" : "NONE");
		printf("DATA DIGEST             : %s\n",
		       (target_details.session_data.
			DataDigest) ? "CRC32" : "NONE");
		printf("FIRSTBURSTLENGTH        : %d\n",
		       target_details.session_data.FirstBurstLength);
		printf("MAXBURSTLENGTH          : %d\n",
		       target_details.session_data.MaxBurstLength);
		printf("MAXRECVDATASEGMENTLENGTH: initiator - "
		       "%d\n",
		       target_details.session_data.MaxRecvDataSegmentLength);
		printf("                          target    - "
		       "%d\n",
		       target_details.session_data.MaxXmitDataSegmentLength);
		printf("LOGIN TIMEOUT           : %d\n",
		       target_details.session_data.login_timeout);
		printf("AUTH TIMEOUT            : %d\n",
		       target_details.session_data.auth_timeout);
		printf("ACTIVE TIMEOUT          : %d\n",
		       target_details.session_data.active_timeout);
		printf("IDLE TIMEOUT            : %d\n",
		       target_details.session_data.idle_timeout);
		printf("PING TIMEOUT            : %d\n",
		       target_details.session_data.ping_timeout);
	    }

	    if (options & LUN_INFO) {

		print_lun_info(target_details, &sinfo_head);
            }
	    printf("************************************************"
		   "*******************************\n");
	    if (options & BUS_SPECIFIC) {
		if (options & TARGET_SPECIFIC)
		    return 0;
		else
		    break;
	    }
	}
	if (options & TARGET_SPECIFIC) {
	    return 0;
	}
    }
    list_iterate(n, &sinfo_head) {
        if(n->path)
            free(n->path);
        if(n->page80)
            free(n->page80);
        if(n->page83_type3)
            free(n->page83_type3);
        if(n->page83_type2)
            free(n->page83_type2);
        if(n->page83_type1)
            free(n->page83_type1);
        if(n->page83_type0)
            free(n->page83_type0);
    }
    list_destroy(&sinfo_head);

    close(ctlfd);
    return 0;
}
